/* Run-time library for use with "p2c", the Pascal to C translator */

/* "p2c"  Copyright (C) 1989 Dave Gillespie.
 * This file may be copied, modified, etc. in any way.  It is not restricted
 * by the licence agreement accompanying p2c itself.
 */

/* To allow to use select */
#include <sys/select.h>

#include <time.h>

#include <utils/p2c.h>


int P_argc;
char **P_argv;

short P_escapecode;
int P_ioresult;
int bigendian;

long EXCP_LINE;    /* Used by Pascal workstation system */

void* __MallocTemp__;

__p2c_jmp_buf *__top_jb;


/* determines endian-ness of machine */
/* contributed by Marwan Jabri, U Sydney */

int isbigendian()
{
	short d = 0x1234;
	char *p = (char *)&d;

	return *p == 0x12;
}


void PASCAL_MAIN(int argc, char **argv)
{
	P_argc = argc;
	P_argv = argv;
	__top_jb = NULL;
	bigendian = isbigendian();
}

/* File functions */

/* Check if at end of file, using Pascal "eof" semantics.  End-of-file for
   stdin is broken; remove the special case for it to be broken in a
   different way. */
int P_eof(FILE *f)
{
    char ch;

	if (feof(f))
		return 1;
	if (f == stdin)
		return 0;    /* not safe to look-ahead on the keyboard! */
	ch = getc(f);
	if (ch == EOF)
		return 1;
	ungetc(ch, f);
		return 0;
}


/* Check if at end of line (or end of entire file). */

int P_eoln(FILE *f)
{
	char ch;

	ch = getc(f);
	if (ch == EOF)
		return 1;
	ungetc(ch, f);
	return (ch == '\n');
}

/* Sets are stored as an array of longs.  S[0] is the size of the set;
   S[N] is the N'th 32-bit chunk of the set.  S[0] equals the maximum
   I such that S[I] is nonzero.  S[0] is zero for an empty set.  Within
   each long, bits are packed from lsb to msb.  The first bit of the
   set is the element with ordinal value 0.  (Thus, for a "set of 5..99",
   the lowest five bits of the first long are unused and always zero.) */

/* (Sets with 32 or fewer elements are normally stored as plain longs.) */

long *P_setunion(long *d, long *s1, long *s2)         /* d := s1 + s2 */
{
	long *dbase = d++;
	long sz1 = *s1++;
	long sz2 = *s2++;

	while (sz1 > 0 && sz2 > 0)
	{
		*d++ = *s1++ | *s2++;
		--sz1;
		--sz2;
	}

	while (--sz1 >= 0)
		*d++ = *s1++;

	while (--sz2 >= 0)
		*d++ = *s2++;

	*dbase = d - dbase - 1;
	return dbase;
}

long *P_setdiff(long *d, long *s1, long *s2)          /* d := s1 - s2 */
{
	long *dbase = d++;
	long sz1 = *s1++;
	long sz2 = *s2++;

	while (--sz1 >= 0 && --sz2 >= 0)
		*d++ = *s1++ & ~*s2++;

	if (sz1 >= 0)
	{
		while (sz1-- >= 0)
			*d++ = *s1++;
	}
	while (--d > dbase && !*d) ;

	*dbase = d - dbase;
	return dbase;
}

int P_inset(unsigned val, long *s)                 /* val IN s */
{
	long bit = val % SETBITS;

	val /= SETBITS;
	return ((long)val < *s++ && ((1<<bit) & s[val]));
}

long *P_addset(long *s, unsigned val)              /* s := s + [val] */
{
	long *sbase = s;
	long bit;
	size_t size;

	bit = val % SETBITS;
	val /= SETBITS;

	size = *s;
	if (++val > size)
	{
		s += size;
		while (val > size)
			*++s = 0, size++;
		*sbase = size;
	}
	else
	{
		s += val;
	}

	*s |= 1 << bit;
	return sbase;
}


long *P_addsetr(long *s, unsigned v1, unsigned v2)              /* s := s + [v1..v2] */
{
	long *sbase = s;
	long b1, b2;
	size_t size;

	if ((int)v1 > (int)v2)
		return sbase;

	b1 = v1 % SETBITS;
	v1 /= SETBITS;
	b2 = v2 % SETBITS;
	v2 /= SETBITS;
	size = *s;
	v1++;
	if (++v2 > size)
	{
		while (v2 > size)
			s[++size] = 0;
		s[v2] = 0;
		*s = v2;
	}
	s += v1;
	if (v1 == v2)
	{
		*s |= (~((unsigned)(-2)<<(b2-b1))) << b1;
    }
	else
	{
		*s++ |= (unsigned)(-1) << b1;
		while (++v1 < v2)
			*s++ = -1;
		*s |= ~((unsigned)(-2) << b2);
    }

	return sbase;
}


long *P_remset(long *s, unsigned val)              /* s := s - [val] */
{
	long bit = val % SETBITS;
	val /= SETBITS;
	if ((long)++val <= *s)
	{
		if (!(s[val] &= ~(1<<bit)))
			while (*s && !s[*s])
				(*s)--;
	}

	return s;
}

int P_subset(long *s1, long *s2)                /* s1 <= s2 */
{
	long sz1 = *s1++;
	long sz2 = *s2++;
	if (sz1 > sz2)
		return 0;

	while (--sz1 >= 0)
	{
		if (*s1++ & ~*s2++)
			return 0;
	}

	return 1;
}


/* s is a "smallset", i.e., a 32-bit or less set stored
   directly in a long. */

long *P_expset(long *d, long s)                /* d := s */
{
	if (s)
	{
		d[1] = s;
		*d = 1;
    }
	else
	{
		*d = 0;
	}

	return d;
}

void* _OutMem()
{
    _Escape(-2);
	return NULL;
}

void _CaseCheck()
{
    _Escape(-9);
}

void _NilCheck()
{
    _Escape(-3);
}

/* The following is suitable for the HP Pascal operating system.
   It might want to be revised when emulating another system. */

char *_ShowEscape(char *buf, int code, int ior, char *prefix)
{
	char *bufp;

	if (prefix && *prefix)
	{
		strcpy(buf, prefix);
		strcat(buf, ": ");
		bufp = buf + strlen(buf);
	}
	else
	{
		bufp = buf;
	}

	if (code == -10)
	{
		sprintf(bufp, "Pascal system I/O error %d", ior);
		switch (ior)
		{
			case 3:
				strcat(buf, " (illegal I/O request)");
				break;
			case 7:
				strcat(buf, " (bad file name)");
				break;
			case FileNotFound:   /*10*/
				strcat(buf, " (file not found)");
				break;
			case FileNotOpen:    /*13*/
				strcat(buf, " (file not open)");
				break;
			case BadInputFormat: /*14*/
				strcat(buf, " (bad input format)");
				break;
			case 24:
				strcat(buf, " (not open for reading)");
				break;
			case 25:
				strcat(buf, " (not open for writing)");
				break;
			case 26:
				strcat(buf, " (not open for direct access)");
				break;
			case 28:
				strcat(buf, " (string subscript out of range)");
				break;
			case EndOfFile:      /*30*/
				strcat(buf, " (end-of-file)");
				break;
			case FileWriteError: /*38*/
				strcat(buf, " (file write error)");
				break;
		}
	}
	else
	{
		sprintf(bufp, "Pascal system error %d", code);
		switch (code)
		{
			case -2:
				strcat(buf, " (out of memory)");
				break;
			case -3:
				strcat(buf, " (reference to NIL pointer)");
				break;
			case -4:
				strcat(buf, " (integer overflow)");
				break;
			case -5:
				strcat(buf, " (divide by zero)");
				break;
			case -6:
				strcat(buf, " (real math overflow)");
				break;
			case -8:
				strcat(buf, " (value range error)");
				break;
			case -9:
				strcat(buf, " (CASE value range error)");
				break;
			case -12:
				strcat(buf, " (bus error)");
				break;
			case -20:
				strcat(buf, " (stopped by user)");
				break;
		}
	}
	return buf;
}

void _Escape(int code)
{
	char buf[100];

	P_escapecode = code;
	if (__top_jb)
	{
		__p2c_jmp_buf *jb = __top_jb;
		__top_jb = jb->next;
		longjmp(jb->jbuf, 1);
	}

	if (code == 0)
		exit(0);
	if (code == -1)
		exit(1);

	fprintf(stderr, "%s\n", _ShowEscape(buf, P_escapecode, P_ioresult, ""));
	exit(1);
}

void _EscIO(int code)
{
	P_ioresult = code;
	_Escape(-10);
}

char *GetChipmunkPath(char *ev, char *deft)
{
	char *s;

	if ((s = getenv(ev)) != (char *) NULL)
		return(s);
    else
		return(deft);
}

/*
    usleep -- support routine for 4.2BSD system call emulations
    last edit:  29-Oct-1984     D A Gwyn
*/
//int select();

int microsleep(long usec)         /* returns 0 if ok, else -1 */
{
    static struct              /* `timeval' */
      {
         long  tv_sec;         /* seconds */
         long  tv_usec;        /* microsecs */

      }   delay;               /* _select() timeout */

    delay.tv_sec = usec / 1000000L;
    delay.tv_usec = usec % 1000000L;

#if defined(__linux__) || defined( __osf__ )
    return select( 0, (fd_set *)0, (fd_set *)0, (fd_set *)0, (struct timeval *)&delay );
#else
    return select( 0, (int *)0, (int *)0, (int *)0, (struct timeval *)&delay );
#endif

}
