#include <utils/p2c.h>
#include <utils/strlist.h>
#include <utils/strings.h>

void strlist_init(strlist_t **base)
{
	*base = NULL;
}

strlist_t *strlist_add(strlist_t **base, char *s)
{
	strlist_t *lp;
	size_t len = strlen(s);

	/* go to good position */
	while ((lp = *base) && strcmp(lp->s, s) < 0)
		base = &lp->next;
	
	if (lp == NULL || strcmp(lp->s, s))
	{
		lp = (strlist_t *)malloc(sizeof(strlist_t));
		if (lp == NULL)
			_OutMem();

		lp->s = (char *)malloc(sizeof(char) * (len + 1));
		if (lp->s == NULL)
			_OutMem();

		lp->next = *base;
		*base = lp;
		strcpy(lp->s, s);
		lp->len = len;
	}

	lp->value = (void *)0;
	lp->kind = 0;
	return lp;
}

strlist_t *strlist_append(strlist_t **base, char *s)
{
	strlist_t *lp;
	size_t len = strlen(s);

	/* go to end of list */
	while ((lp = *base))
		base = &lp->next;
	
	lp = (strlist_t *)malloc(sizeof(strlist_t));
	if (!lp)
		_OutMem();
	
	lp->s = (char *)malloc(sizeof(char) * (len + 1));
	if (lp->s == NULL)
		_OutMem();

	lp->next = NULL;
	*base = lp;
	strcpy(lp->s, s);
	lp->len = len;
	lp->value = (void *)0;
	lp->kind = 0;
	return lp;
}

strlist_t *strlist_insert(strlist_t **base, char *s)
{
	strlist_t *lp;
	size_t len = strlen(s);

	lp = (strlist_t *)malloc(sizeof(strlist_t));
	if (!lp)
		_OutMem();
	lp->s = (char *)malloc(sizeof(char) * (len + 1));
	if (lp->s == NULL)
		_OutMem();
	
	lp->next = *base;
	*base = lp;
	strcpy(lp->s, s);
	lp->len = len;
	lp->value = (void *)0;
	lp->kind = 0;
	return lp;
}

strlist_t *strlist_find(strlist_t *base, char *s)
{
	while (base && strcmp(base->s, s))
		base = base->next;
	return base;
}

strlist_t *strlist_cifind(strlist_t *base, char *s)
{
	while (base && strcicmp(base->s, s))
		base = base->next;
	return base;
}

void strlist_change(strlist_t **base, strlist_t **lp, char *s)
{
	strlist_t *lp2 = *lp;
	strlist_t *lp3;
	size_t len = strlen(s);

	while ((lp3 = *base) != lp2)
	{
		if (lp3)
			base = &lp3->next;
		else
			return;   /* wasn't in the list */
	}
	
	/* must reallocate */
	if (len > lp3->len)
	{
		lp3->s = (char *)realloc(lp3->s,  (len + 1) * sizeof(char));
		if (lp3->s == NULL)
			_OutMem();
		
		lp3->len = len;
		*lp = lp3;
		lp3->next = *base;
		*base = lp3;
	}

	strcpy(lp3->s, s);
}

void strlist_remove(strlist_t **base, char *s)
{
	strlist_t *lp;

	while ((lp = *base) != NULL)
	{
		if (!strcmp(lp->s, s))
		{
			*base = lp->next;
			strlist_dispose(&lp);
			return;
		}
		base = &lp->next;
	}
}

void strlist_delete(strlist_t **base, strlist_t **lp)
{
	strlist_t *lp2;

	while ((lp2 = *base) != NULL)
	{
		if (lp2 == *lp)
		{
			*base = (*lp)->next;
			strlist_dispose(lp);
			return;
		}
		base = &lp2->next;
	}
}

void strlist_dispose(strlist_t **ptr)
{
	free((*ptr)->s);
	free(*ptr);
	*ptr = NULL;
}

void strlist_empty(strlist_t **base)
{
	strlist_t *lp;

	while ((lp = *base) != NULL)
	{
		*base = lp->next;
		strlist_dispose(&lp);
	}
	*base = NULL;
}

void strlist_copy(strlist_t **dest, strlist_t *src)
{
	strlist_t *lcopy;

	strlist_empty(dest);

	while (src != NULL)
	{
		lcopy = (strlist_t *)malloc(sizeof(strlist_t));
		if(lcopy == NULL)
			_OutMem();
		lcopy->s = (char *)malloc(sizeof(char) * (src->len + 1));

		lcopy->len = src->len;
		lcopy->kind = src->kind;
		lcopy->value = src->value;
		strcpy(lcopy->s, src->s);

		*dest = lcopy;

		/* move to next node */
		dest = &lcopy->next;
		src = src->next;
	}

	*dest = NULL;
}

void strlist_dump(FILE *f, strlist_t *base)
{
	while (base != NULL)
	{
		fprintf(f, "\"%s\" %d, %ld\n", base->s, base->kind, (long)base->value);
		base = base->next;
	}
}

